/* Utility routines for remote SunOS kernel debugging.
   Copyright (C) 1986, 1989 Free Software Foundation, Inc.
   Copyright (C) 1992,1993,1994 by John Heidemann <johnh@cs.ucla.edu>

   This file is derived from rem-multi/remote_utils.c from gdb 4.5.

	@(#)kgdb_utils.c	1.13 (John Heidemann) 9/12/94

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "defs.h"
#include <debug/debugger.h>
#include "kgdb.h"
#include "kgdb_gdb.h"


/*
 * Serial chip stuff.
 */
#ifdef DIRECT_DRIVE

#if RECOVER_COUNT == 0
#define RECOVER()
#else /* RECOVER_COUNT */
#define RECOVER()	do { \
			    int rec_ctr = RECOVER_COUNT; \
			    while (--rec_ctr >= 0) \
				    ; \
			} while (0)
#endif /* RECOVER_COUNT */

#define CSR(x)	(x)
#define DR(x)	((x) + 2)
#define IRDY	0x01		/* Bit that indicates input is ready */
#define ORDY	0x02		/* Bit that indicates output is ready */

/*
 * The following values are the actual base addresses of the serial
 * chips in I/O space.  If you don't know what the correct values are,
 * and kgdb doesn't work, turn off DIRECT_DRIVE and turn on
 * DUMP_DESCRIPTORS.  Look through the output for something that looks
 * like an I/O base (it will probably be the third longword of the dump).
 */
char *remote_base = (char *) REMOTE_BASE; /* Patchable */
char *stdin_base = (char *) STDIN_BASE; /* Patchable */
#endif /* DIRECT_DRIVE */



/*
 * A spinner slows things down a lot, but it lets you know
 * you're not hung.  This should be off if you're doing DIRECT_DRIVE
 * (it defeats the purpose).
 */
#ifdef USE_SPINNER
#define SPINNER_DEFAULT 1
#else
#define SPINNER_DEFAULT 0
#endif /* USE_SPINNER */
int kiospinner = SPINNER_DEFAULT; /* flag to enable the spinning bar when IOing */

#ifdef USE_KIODEBUG
#define KIODEBUG_DEFAULT 1
#else
#define KIODEBUG_DEFAULT 0
#endif /* USE_KIODEBUG */
int kiodebug = KIODEBUG_DEFAULT;		/* flag to enable verbose debugging */

/*
 * The kioecho serial i/o protocol requires that we echo each character
 * back that we get.  It's stupid and slow, but it helps if you
 * drop random characters during fast transmission.
 */
#ifdef USE_KIOECHO
#define KIOECHO_DEFAULT 1
#else
#define KIOECHO_DEFAULT 0
#endif /* USE_KIOECHO */
int kioecho = KIOECHO_DEFAULT;	/* flag to enable the echo-characters-back protocol */

/* patchable -- requires openprom v2.0 or better */
char kgdb_device[100] = KGDB_DEVICE;


#define	PBUFSIZ	1024

ihandle_t remote_desc = (ihandle_t)NULL;
ihandle_t stdin_desc = (ihandle_t)NULL;
int remote_debugging = 0;	/* flag set if we successfully init ourselves */

int timeout = 10;

int inferior_pid = 0;
int ignore_checksum = 0; /* patchable */





/*
 * Some random user-level functions.
 */
void
perror(s)
	const char *s;
{
	extern int errno;

	printf ("%s: errno=%d\n", s, errno);
}


void
panic(s)
	char *s;
{
	printf ("panic: %s\n", s);
	/* NEEDSWORK - should exit */
}



/*
 * Spinner code.  Let a watcher know we're up.
 */

void
spinner()
{
#define NBARS 4
	static char bars[] = "-\\|/";
	static int n=0;

	if (!kiospinner)
		return;
	printf ("\b%c", bars[n++ % NBARS]);
}

void
prespinner()
{
	if (!kiospinner)
		return;
	printf (" ");
	spinner();
}

void
postspinner()
{
	if (!kiospinner)
		return;
	printf ("\b \b");
}



/* Open a connection to a remote debugger.
   NAME is the filename used for communication.  */

/*
 * Communications paramters are set by the eeprom variable
 * ttyb-mode.  I haven't been able to get it working faster than 9600 baud.
 * If you succeed, let me know.
 */

void
remote_open ()
{
	char *name = kgdb_device;
	remote_debugging = 0;

	if (romp->v_romvec_version == 0) {
		for (;;) {
			printf ("kgdb: kgdb requires PROM version 2.0 or later.");
			montrap();
		};
	};

	if ((ihandle_t)NULL == (remote_desc = ((*romp->op_open)(name))))
		panic ("remote_open: cannot open remote port\n");

	stdin_desc = *romp->op_stdin;

	if (kiodebug)
		printf ("Remote debugging using %s\n", name);

	inferior_pid = 0x1;
	remote_debugging = 1;
}

void
remote_close()
{
	printf ("Remote debugging port closed.\n");
	(*romp->op_close)(remote_desc);

	remote_debugging = 0;

	printf ("Exiting debugger...\n");
	montrap();
	/*
	 * This should never be reached, but what if they type "go"?
	 * Just in case, try and restart things.
	 */
	remote_open();
}


/* Convert hex digit A to a number.  */

static int
fromhex (a)
     int a;
{
	if (a >= '0' && a <= '9')
		return a - '0';
	else if (a >= 'a' && a <= 'f')
		return a - 'a' + 10;
	else
		perror ("Reply contains invalid hex digit");
	return 0;
}

/* Convert number NIB to a hex digit.  */

static int
tohex (nib)
     int nib;
{
	if (nib < 10)
		return '0'+nib;
	else
		return 'a'+nib-10;
}


/* Read/write a single character from/to the remote end. */

static void
writechars(buf, len)
	char *buf;
	int len;
{
#if defined(DIRECT_DRIVE) && 0
	/*
	 * This seems to be hanging, so it's if'ed out
	 * pending Geoff's review.
	 */
	while (--len >= 0) {
		while ((*CSR(remote_base) & ORDY) == 0) {
			kiospinner();
		};
		*DR(remote_base) = *buf++;
	}

#else /* DIRECT_DRIVE */

	int n;
	n = (*romp->op_write) (remote_desc, buf, len);
	if (n != len)
		printf ("kgdb writechars: op_write says %d != %d\n",
			len, n);
#endif /* DIRECT_DRIVE */
}

static int
poll_console()
{
	int n;
	char ch;

	/* op_read is non-blocking */
	n = (*romp->op_read)(stdin_desc, &ch, 1);
 	if (kiodebug)
		printf("kgdb: poll_console, ch=%d, n=%d\n", ch, n);
	if (n == 1) {
		switch (ch) {
		case 3:
			/*
			 * Check for ^C's on the
			 * console to abort.  We could
			 * do _longjmp(abort_jmp, 1),
			 * but let's just drop into
			 * the monitor instead.
			 */
			printf ("^C");
			montrap();
			break;
		case 17:
			/*
			 * Check for ^Q (analgous to XON)
			 * which says ``just return''.
			 * Currently the return code
			 * is kind of a hack---we
			 * return a ``character'' of
			 * -1 and check for it in some
			 * (but not all :-( ) places.
			 */
			if (!kiospinner)
				printf ("^Q");
			return -1;
		case 8:
			/*
			 * Control-i is yet another hack:
			 * toggle checksum ignore.
			 */
			ignore_checksum = !ignore_checksum;
			printf ("kgdb: checsums %s ignored\n",
				ignore_checksum ? "are" : "are not");
			break;
		default:	/* do nothing otherwise */
			break;
		}
	}
	return 0;
}	

#ifdef DUMP_DESCRIPTORS
static void
dump_descriptors()
{
	static int firsttime = 1;
	int i;

	if (firsttime) {
		firsttime = 0;
		printf("remote_desc:  0x%x\n", remote_desc);
		for (i = 0;  i < 256;  i += 4) {
			printf("0x%x ", *(long *)((char *)remote_desc + i));
		}
		printf("\n");
		printf("stdin_desc:  0x%x\n", stdin_desc);
		for (i = 0;  i < 256;  i += 4) {
			printf("0x%x ", *(long *)((char *)stdin_desc + i));
		}
		printf("\n");
	}
}
#else
#define dump_descriptors() /*nothing*/
#endif

static int
readchar()
{
#if defined(DIRECT_DRIVE)
	char ch;
	int n;
	int pollcount;
	int e;

	dump_descriptors();

	for (pollcount = POLLCOUNT;;) {
		RECOVER();
		if (*CSR(remote_base) & IRDY) {
			RECOVER();
			ch = *DR(remote_base);
			n = 1;
		} else {
			n = 0;
		}
		if (n == 1) {
			if (kioecho)
				writechars(&ch, 1);
			break;
		} else if (--pollcount <= 0) {
			pollcount = POLLCOUNT;
			if (e = poll_console())
				return e;
		}
	}
	return ch;

#elif defined(NO_HARDWARE_HANDSHAKE)

	char ch;
	int n;
	int pollcount;
	int e;

	for (pollcount = POLLCOUNT;;) {
		n = (*romp->op_read)(remote_desc, &ch, 1);
		if (n == 1) {
			if (kioecho)
				writechars(&ch, 1);
			break;
		} else if (--pollcount <= 0) {
			pollcount = POLLCOUNT;
			if (e = poll_console())
				return e;
		}
	}
	return ch;

#else /* ! (defined(NO_HARDWARE_HANDSHAKE) || defined(DIRECT_DRIVE)) */

	/*
	 * The GNU version.
	 */
	static int inbuf_index, inbuf_count;
	static char inbuf[PBUFSIZ];

	if (inbuf_index >= inbuf_count)
	{
		/* Time to do another read... */
		inbuf_index = 0;
		inbuf_count = 0;
		inbuf[0] = 0;		/* Just in case */

		/* termio does the timeout for us.  */
		inbuf_count = read (remote_desc, inbuf, INBUFSIZE);
		if (inbuf_count == 0)
			return 0;   /* 0 = time out */
	}

	/* Just return the next character from the buffer.  */
	return inbuf[inbuf_index++] & 0x7f;
#endif
}





/* Send a packet to the remote machine, with error checking.
   The data of the packet is in BUF.  */

void
putpkt (buf)
	char *buf;
{
	int i;
	unsigned char csum = 0;
	char buf2[PBUFSIZ];
	int cnt = strlen (buf);
	char *p;
	int ch;

	/* Copy the packet into buffer BUF2, encapsulating it
	   and giving it a checksum.  */

	if (cnt > sizeof(buf2) - 5)		/* Prosanity check */
		abort();

	p = buf2;
	*p++ = '$';

	for (i = 0; i < cnt; i++) {
		csum += buf[i];
		*p++ = buf[i];
	}
	*p++ = '#';
	*p++ = tohex ((csum >> 4) & 0xf);
	*p++ = tohex (csum & 0xf);
	*p++ = '\n';  /* JSH hack */

	/* Send it over and over until we get a positive ack.  */

	do {
		if (kiodebug) {
			*p = '\0';
			printf ("Sending packet: %s...", buf2);
		}
		else spinner();
		cnt = p-buf2;
		writechars (buf2, cnt);

		/* read until either a timeout occurs (\0) or '+' is read */
		do {
			ch = readchar();
			if (ch == 3) {
				/*
				 * Hmm.  We got an interrupt.
				 * The other guy thinks we shouldn't
				 * be doing what we are.  How to recover?
				 * Just bail.  Assume an ack.
				 */
				printf ("kgdb: putpkt interrupted.\n");
				ch = '+';
			};
			if (kiodebug) {
				if (ch == 3)
					printf("Int\n");
				else if (ch == '+')
					printf("Ack\n");
				else if (ch == '-') {
					printf("Nak\n");
					ch = 0;
				} else printf ("%2x%c ", (ch & 0xff), ch);
			}
		} while ((ch != '+') && (ch != '\0'));
	} while (ch != '+');
}

static void
create_continue_pkt(buf)
	char *buf;
{
	buf[0] = 'C';
	buf[1] = 0;
}

static void
create_interrupt_pkt(buf)
	char *buf;
{
	buf[0] = 3;
	buf[1] = 0;
}

/* Read a packet from the remote machine, with error checking,
   and store it in BUF.  */

void
getpkt(buf)
	char *buf;
{
	char *bp = buf;
	unsigned char csum, sentsum;
	int c;
	unsigned char c1, c2;
#define MODE_BEFORE_PKT 1
#define MODE_IN_PKT 2
#define MODE_IN_CS1 3
#define MODE_IN_CS2 4
#define MODE_AFTER_PKT 5
	int mode;
#define DEBUG_STATE_TRANSITION(STATE) if (kiodebug) printf ("kgdb: getpkt: %s\n", STATE);

	while (1) {
		/* Force csum to be zero here because of possible error retry.  */
		csum = 0;

		mode = MODE_BEFORE_PKT;
		while (mode != MODE_AFTER_PKT) {
			c = readchar();
			if (c == -1) {
				/*
				 * -1 flag (pesudo-EOF) indicates
				 * the user has requested us to continue.
				 * Fake up a "C"ontinue packet
				 * to return to kgdb_loop.
				 * (Note that an upper-case C packet
				 * is handled slightly differently 
				 * than a lower case c packet;
				 * see kgdb_loop for details.)
				 */
				create_continue_pkt(buf);
				return;
			} else if (c == 3) {
				/*
				 * If we get a ctrl-C OTW,
				 * indicate this and return.
				 * gdb wants to get a packet back.
				 * Send kgdb_loop a "^c" pkt
				 * pkt to request status.
				 */
				printf ("kgdb: ctrl-C OTW\n");
				create_interrupt_pkt(buf);
				return;
			} else if (mode == MODE_BEFORE_PKT && c == '$') {
				bp = buf;
				csum = 0;
				mode = MODE_IN_PKT;
				DEBUG_STATE_TRANSITION("in_pkt");
			} else if (mode == MODE_IN_PKT && c == '#') {
				mode = MODE_IN_CS1;
				/* DEBUG_STATE_TRANSITION("in_cs1"); -- too slow */
			} else if (mode == MODE_IN_PKT) {
				*bp++ = c;
				csum += c;
			} else if (mode == MODE_IN_CS1) {
				*bp = 0;
				c1 = fromhex(c);
				mode = MODE_IN_CS2;
				/* DEBUG_STATE_TRANSITION("in_cs2"); -- too slow */
			} else if (mode == MODE_IN_CS2) {
				c2 = fromhex(c);
				mode = MODE_AFTER_PKT;
				DEBUG_STATE_TRANSITION("after_pkt");
			};
			if (kiodebug)
				printf ("%2x%c ", (c & 0xff), c);
			else spinner();
		}

		sentsum = (c1 << 4) + c2;
		if ((csum & 0xff) == sentsum)
			break;
		printf ("Bad checksum, sentsum=0x%x, csum=0x%x, buf=%s\fn",
			(sentsum & 0xff), (csum & 0xff), buf);
		if (ignore_checksum) {
			printf ("Bad checksum ignored.\n");
			break;
		};
		writechars ("-", 1);
	}

	writechars ("+", 1);

	if (kiodebug)
		printf ("Packet received: %s\n", buf);
}



void
write_ok(buf)
	char *buf;
{
	buf[0] = 'O';
	buf[1] = 'K';
	buf[2] = '\0';
}

void
write_enn(buf)
	char *buf;
{
	buf[0] = 'E';
	buf[1] = 'N';
	buf[2] = 'N';
	buf[3] = '\0';
}

void
convert_int_to_ascii(from,to,n)
	char *from, *to;
	int n;
{
 	int nib;
	char ch;
	while(n--) {
		ch = *from++;
		nib = ((ch & 0xf0) >> 4)& 0x0f;
		*to++ = tohex(nib);
		nib = ch & 0x0f;
		*to++ = tohex(nib);
	}
	*to++ = 0;
}


void
convert_ascii_to_int(from,to,n)
	char *from, *to;
	int n;
{
 	int nib1,nib2 ;
	while( n-- )
	{
		nib1 = fromhex(*from++);
		nib2 = fromhex(*from++);
		*to++ = (((nib1 & 0x0f)<< 4)& 0xf0) | (nib2 & 0x0f);
	}
}

void
prepare_resume_reply(buf, signal)
	char *buf;
	unsigned char signal;
{
 	int nib;

	*buf++ = 'S';
	nib = ((signal & 0xf0) >> 4) ;
	*buf++ = tohex(nib);
	nib = signal & 0x0f;
	*buf++ = tohex(nib);
	*buf++ = 0;
}

static char *
bufify_register(buf, regno, reg, regsize)
	char *buf;
	int regno;
	char *reg;
	int regsize;
{
	char ch = regno;
	convert_int_to_ascii(&ch, buf, sizeof(char));
	buf += 2 * sizeof(char);
	*buf++ = ':';
	convert_int_to_ascii(reg, buf, regsize);
	buf += regsize * 2;
	*buf++ = ';';
	return buf;
}

void
prepare_long_resume_reply(buf, signal, registers)
	char *buf;
	unsigned char signal;
	char *registers;
{
 	int nib;

	*buf++ = 'T';
	nib = ((signal & 0xf0) >> 4) ;
	*buf++ = tohex(nib);
	nib = signal & 0x0f;
	*buf++ = tohex(nib);
	buf = bufify_register(buf, PC_REGNUM,
			      &registers[REGISTER_BYTE(PC_REGNUM)], KGDB_REGSIZE);
	buf = bufify_register(buf, SP_REGNUM,
			      &registers[REGISTER_BYTE(SP_REGNUM)], KGDB_REGSIZE);
	buf = bufify_register(buf, FP_REGNUM,
			      &registers[REGISTER_BYTE(FP_REGNUM)], KGDB_REGSIZE);
	*buf++ = 0;
}

void
decode_m_packet(from,mem_addr_ptr,len_ptr)
	char *from;
	unsigned int *mem_addr_ptr, *len_ptr;
{
	int i = 0, j = 0 ;
	char ch;
	*mem_addr_ptr = *len_ptr = 0;

	while((ch = from[i++]) != ',') {
		*mem_addr_ptr = *mem_addr_ptr << 4;
		*mem_addr_ptr |= fromhex(ch) & 0x0f;
	}

	for(j=0; j < 4; j++) {
		if((ch = from[i++]) == 0)
			break;
		*len_ptr = *len_ptr << 4;
		*len_ptr |= fromhex(ch) & 0x0f;
	}
}

void
decode_M_packet(from,mem_addr_ptr,len_ptr,to)
	char *from, *to;
	unsigned int *mem_addr_ptr, *len_ptr;
{
	int i = 0;
	char ch;
	*mem_addr_ptr = *len_ptr = 0;

	while((ch = from[i++]) != ',') {
		*mem_addr_ptr = *mem_addr_ptr << 4;
		*mem_addr_ptr |= fromhex(ch) & 0x0f;
	}

	while((ch = from[i++]) != ':') {
		*len_ptr = *len_ptr << 4;
		*len_ptr |= fromhex(ch) & 0x0f;
	}

	convert_ascii_to_int(&from[i++],to,*len_ptr);
}

